import math
from functools import reduce


def float_gcd(*numbers: float) -> float:
    """Calculate the greatest common divisor of a list of floats.

    Example:
        >>> float_gcd(0.75, 0.25, 0.375)
        0.125

    Args:
        numbers (float): A list of floats.

    Returns:
        float: The greatest common divisor of the list of floats.
    """
    return reduce(_gcd, numbers)


def _gcd(a: float, b: float) -> float:
    # Make sure that a >= b
    if a < b:
        return _gcd(b, a)

    # If b is close to 0, then a is the answer
    if abs(b) < 0.001:
        return a
    # If b is not close to 0, then the answer is the GCD of b and the remainder of a/b
    else:
        return _gcd(b, a - math.floor(a / b) * b)


if __name__ == "__main__":
    print(float_gcd(0.75, 0.25, 0.375))
    print(float_gcd(15, 20))



"""Module to construct, simulate, and visualize quantum circuit"""

from __future__ import annotations

import itertools
import typing
from typing import List

import numpy as np
from IPython.display import Latex
from numpy.typing import NDArray
from qiskit import ClassicalRegister, QuantumCircuit, QuantumRegister, quantum_info, transpile
from qiskit.visualization import plot_histogram
from qiskit_aer import AerSimulator

from qiskit_extension import latex_drawer
from qiskit_extension.state_vector2 import StateVector2

if typing.TYPE_CHECKING:
    import matplotlib.figure


class QuantumCircuit2(QuantumCircuit):
    """An extended class of QuantumCircuit from Qiskit:
    https://qiskit.org/documentation/stubs/qiskit.circuit.QuantumCircuit.html

    Args:
        number_of_qubits (int, optional): Number of qubits. Defaults to 1.
    """

    def __init__(self, number_of_qubits: int = 1):
        # Use Aer's qasm_simulator
        self.simulator = AerSimulator()
        self.n = number_of_qubits
        q = QuantumRegister(self.n, "q")
        c = ClassicalRegister(self.n, "c")
        super().__init__(q, c)

    def to_matrix(self) -> NDArray[np.complex128]:
        """Return matrix form of the quantum circuit"""
        reverse_qc = self.reverse_bits()  # REVERSE the order of qubits to fit textbook notation
        return quantum_info.Operator(reverse_qc).data

    def show_matrix(self):
        """Show the matrix form of the quantum circuit in LaTeX"""
        return latex_drawer.matrix_to_latex(self.to_matrix())

    def show_circ(self) -> Latex:
        """Show the circuit in LaTeX"""
        return self.draw(output="latex", idle_wires=False)  # type: ignore

    def show_measure_all(self) -> matplotlib.figure.Figure:
        """Measure every qubit at the end of the circuit, then plot the result

        Returns:
            matplotlib.figure.Figure: A histogram of the measurement result
        """
        for i in range(self.n):
            self.measure(i, i)

        # compile the circuit down to low-level QASM instructions
        # supported by the backend (not needed for simple circuits)
        compiled_circuit = transpile(self, self.simulator)

        # Execute the circuit on the qasm simulator
        job = self.simulator.run(compiled_circuit, shots=1000)

        # Grab results from the job
        result = job.result()

        # Returns counts
        counts = result.get_counts(self)
        print(f"\nTotal count : {counts}\n")

        # Plot a histogram
        return plot_histogram(counts)

    def get_state(self) -> StateVector2:
        """Initialize the state to ground state, then evolve the state by the quantum circuit.

        Returns:
            StateVector2: Final state of the quantum circuit
        """
        # Set the initial state of the simulator to the ground state using from_int
        state = StateVector2.from_int(i=0, dims=2**self.n)

        # Evolve the state by the quantum circuit
        state = state.evolve(self)
        return state

    def phase_shift(self, theta, qubit) -> None:
        """Phase shift gate

        Args:
            theta (_type_): phase shift angle
            qubit (_type_): which qubit to apply the gate
        """
        self.p(theta, qubit)
        self.x(qubit)
        self.p(theta, qubit)
        self.x(qubit)

    def h_all(self) -> None:
        """Perform H gate on every qubit"""
        for i in range(self.n):
            self.h(i)

    def cz_line_all(self) -> None:
        """Perform CZ gate on qubit(0,1), qubit(1,2), qubit(2,3) ... , qubit(n-1,n)"""
        for i in range(self.n - 1):
            self.cz(i, i + 1)

    def cz_star_all(self, center=0) -> None:
        """Perform CZ gate on (center qubit, every other qubit)

        Args:
            center (int, optional): Center qubit. Defaults to 0.
        """
        for i in range(self.n):
            if i == center:
                continue
            self.cz(center, i)

    def cz_complete(self) -> None:
        """Perform CZ gate on every possible combination of qubits"""
        for qubit_c, qubit_t in itertools.combinations(range(self.n), 2):
            self.cz(qubit_c, qubit_t)

    def z2epr_pair(self, qubit_c: List[int] | int, qubit_t: List[int] | int) -> None:
        """Perform Z to EPR pair transformation

        Args:
            qubit_c (List[int] | int): control qubit
            qubit_t (List[int] | int): target qubit
        """
        if isinstance(qubit_c, int) and isinstance(qubit_t, int):
            qubit_c, qubit_t = [qubit_c], [qubit_t]
        if isinstance(qubit_c, list) and isinstance(qubit_t, list):
            if len(qubit_c) != len(qubit_t):
                raise ValueError("qubit_c and qubit_t must have the same length")
            for c, t in zip(qubit_c, qubit_t):
                self.h(c)
                self.cx(c, t)
        else:
            raise TypeError("qubit_c and qubit_t must be List[int] or int")

    def epr2z_basis(self, qubit_c: List[int] | int, qubit_t: List[int] | int) -> None:
        """Perform EPR pair to Z basis transformation

        Args:
            qubit_c (List[int] | int): control qubit
            qubit_t (List[int] | int): target qubit
        """
        if isinstance(qubit_c, int) and isinstance(qubit_t, int):
            qubit_c, qubit_t = [qubit_c], [qubit_t]
        if isinstance(qubit_c, list) and isinstance(qubit_t, list):
            if len(qubit_c) != len(qubit_t):
                raise ValueError("qubit_c and qubit_t must have the same length")
            for c, t in zip(qubit_c, qubit_t):
                self.cx(c, t)
                self.h(c)
        else:
            raise TypeError("qubit_c and qubit_t must both be List[int] or both be int")



from functools import reduce

import numpy as np


def tensor_product(*matrices: np.ndarray) -> np.ndarray:
    """Tensor product of a list of square matrix or vector."""
    return reduce(np.kron, matrices)


def inverse_tensor(array: np.ndarray) -> np.ndarray:
    """Inverse tensor product of a square matrix or vector.
    Only accept square matrix or vector with length of 2^n.

    Example:
        input  = A ⊗ B ⊗ C ,
        output = C ⊗ B ⊗ A ,
        (A, B, C are square matrix or vector with length of 2)

    Args:
        array (ndarray): A square matrix or vector.

    Raises:
        ValueError: If the input is not a square matrix or vector.
        ValueError: If the length of input matrix/vector is not 2^n.

    Returns:
        ndarray: The inverse tensor product of input square matrix/vector.
    """
    # If input is a 1d array, extend it to a 2d array
    extend_flag = False
    if len(array.shape) == 1:
        extend_flag = True
        array = array[np.newaxis]

    # Check input validity
    dim_i = array.shape[0]
    dim_j = array.shape[1]
    if dim_i != 1 and dim_j != 1 and dim_i != dim_j:
        raise ValueError("Only accept square matrix or vector.")
    if dim_i & (dim_i - 1) or dim_j & (dim_j - 1):
        raise ValueError("Only accept matrix/vector with length of 2^n.")

    # Linear transform A⊗B⊗...⊗N -> N⊗...⊗B⊗A
    reorder_array = np.zeros_like(array)
    bit_i = int(np.log2(dim_i))
    bit_j = int(np.log2(dim_j))
    for i in range(dim_i):
        # Inverse index (binary), e.g. 1010 -> 0101; 0100 -> 0010
        str_i = bin(i)[2:]
        str_i = "0" * (bit_i - len(str_i)) + str_i
        inverse_i = int(str_i[::-1], 2)
        for j in range(dim_j):
            str_j = bin(j)[2:]
            str_j = "0" * (bit_j - len(str_j)) + str_j
            inverse_j = int(str_j[::-1], 2)
            # Use inverse index to reorder the array
            reorder_array[i, j] = array[inverse_i, inverse_j]

    # If input is a 1d array, squeeze output back to a 1d array
    if extend_flag:
        return reorder_array.squeeze()
    return reorder_array


if __name__ == "__main__":
    array = np.array([[0], [1], [2], [3]])
    print(inverse_tensor(array), "\n")
    array = np.array([[0, 1, 2, 3], [4, 5, 6, 7], [8, 9, 10, 11], [12, 13, 14, 15]])
    print(inverse_tensor(array), "\n")
    array = np.array([0, 1, 2, 3])
    print(inverse_tensor(array), "\n")
    array = np.array([[0, 1, 2, 3]])
    print(inverse_tensor(array), "\n")



"""Module to convert various data types to LaTeX representation.
IPython.display.Latex(KaTeX) Supported Functions: https://katex.org/docs/supported.html
"""

from __future__ import annotations

import math
import os  # use by latex to image
import subprocess  # use by latex to image
import tempfile  # use by latex to image
import typing
from typing import List, Tuple

import numpy as np
import sympy
from numpy.typing import NDArray
from PIL import Image  # use by latex to image
from qiskit.visualization.exceptions import VisualizationError  # use by latex to image
from qiskit.visualization.utils import _trim as trim_image  # use by latex to image

from qiskit_extension.ket import Ket
from qiskit_extension.utils.find_nth_substring import find_nth_substring
from qiskit_extension.utils.float_gcd import float_gcd

if typing.TYPE_CHECKING:
    from qiskit_extension.state_vector2 import StateVector2


def matrix_to_latex(matrix: NDArray[np.float128 | np.complex128]) -> Image.Image:
    """Convert matrix to latex representation

    Args:
        matrix (NDArray[np.float128]): The matrix to be converted

    Returns:
        Image.Image: LaTeX representation of the matrix
    """
    prefix = R"$\begin{bmatrix}"
    suffix = R"\end{bmatrix}$"

    # If input is a 1d array, convert it to a column matrix
    if len(matrix.shape) == 1:
        matrix = matrix[np.newaxis].T

    # Extract the common factor from the matrix
    gcd = float_gcd(*np.absolute(matrix.flatten()))
    if not (math.isclose(gcd, 1) or math.isclose(gcd, 0)):
        matrix = matrix / gcd  # type: ignore
        pretty_gcd = _num_to_latex_ket(gcd)
        prefix = prefix[:1] + pretty_gcd + prefix[1:]

    # Convert the matrix to latex code
    latex_list: List[str] = []
    for i in range(matrix.shape[0]):
        for j in range(matrix.shape[1]):
            # Convert the value to latex code
            pretty_valve = _num_to_latex_ket(matrix[i][j])
            latex_list.append(pretty_valve)
            # Add line up symbol
            if j != matrix.shape[1] - 1:
                latex_list.append("&")
        # Add line break symbol at the end of each row
        latex_list.append(R"\\[6pt]")

    latex_code = prefix + "".join(latex_list) + suffix
    return _latex_code_to_img(latex_code)


def state_to_latex(
    state: StateVector2, hide: List[int] | str = [], output_length: int = 2, source: bool = False
) -> Image.Image | str:
    """Convert state vector to latex representation

    Args:
        state (StateVector2):state vector
        hide (List[int] | str, optional): hide qubits. Default to show all qubits.
        output_length (int, optional): 2^output_length = number of terms in each line. Defaults to 2(= 4 terms/line).
        source (bool, optional): If True, return the latex source code. Defaults to False.

    Returns:
        Image.Image: LaTeX representation of the state vector
    """
    prefix = R"$\begin{alignedat}{" + f"{2**(output_length+1)+1}" + R"}&\; \;&\;"
    suffix = R"\end{alignedat}$"
    latex_code = _state_to_latex_ket(state, hide)
    latex_code = _latex_line_break(latex_code, output_length)
    latex_code = prefix + latex_code + suffix
    latex_code = latex_code.replace(R"\;&\;-", R"-&\;")
    if source:
        return latex_code
    return _latex_code_to_img(latex_code)


def measure_result_to_latex(
    result: Tuple[List[StateVector2], List[StateVector2]],
    hide: List[int] | str = [],
    output_length: int = 2,
    source: bool = False,
) -> Image.Image | str:
    """Convert measurement result to latex representation

    Args:
        result (Tuple[StateVector2, StateVector2]): measurement result
        hide (List[int] | str, optional): hide qubits. Default show all qubits.
        output_length (int, optional): 2^output_length = number of terms in each line. Defaults to 2(= 4 terms/line).
        source (bool, optional): If True, return the latex source code. Defaults to False.

    Returns:
        Latex: LaTeX representation of the measurement result
    """
    prefix = R"$\begin{alignedat}{" + f"{2**(output_length+1)+1}" + "}"
    suffix = R"\end{alignedat}$"

    measure_state_list, remain_state_list = result
    latex_list: List[str] = []
    for measure_state, remain_state in zip(measure_state_list, remain_state_list):
        if measure_state is not None:
            latex_list.append(_state_to_latex_ket(state=measure_state)[:-1])
            latex_list.append(R":&\; \;&\;")
            tmp_str = _state_to_latex_ket(state=remain_state, hide=hide)
            latex_list.append(_latex_line_break(tmp_str, output_length))
            latex_list.append(R"\\\\")
    latex_code = prefix + "".join(latex_list) + suffix
    latex_code = latex_code.replace(R"\;&\;-", R"-&\;")
    if source:
        return latex_code
    return _latex_code_to_img(latex_code)


def _state_to_latex_ket(state: StateVector2, hide: List[int] | str = []) -> str:
    """Convert state vector to latex representation, modified from qiskit.visualization.state_visualization

    Args:
        data: State vector
        hide (List[int] | str, optional): hide qubits. Default show all qubits.

    Returns:
        str: String with LaTeX representation of the state vector
    """
    data = state.data
    basis = state.basis
    num = state._num_of_qubit
    if isinstance(hide, str):
        hide = [int(char) for char in hide]
    for i in hide:
        basis[i] = "hide"

    def ket_name(i):
        ket = bin(i)[2:].zfill(num)
        # ket = ket[::-1]  # REVERSE the order of qubits to fit textbook notation
        new_ket = ""
        for b, k in zip(basis, ket):
            match b:
                case "hide":
                    continue
                case "z":
                    new_ket += Ket.z1 if int(k) else Ket.z0
                case "x":
                    new_ket += Ket.x1 if int(k) else Ket.x0
                case "y":
                    new_ket += Ket.y1 if int(k) else Ket.y0
        return new_ket

    data = np.around(data, 15)
    nonzero_indices = np.where(data != 0)[0].tolist()
    latex_terms = _coeffs_to_latex_terms(data[nonzero_indices], decimals=15)

    latex_list: List[str] = []
    for idx, ket_idx in enumerate(nonzero_indices):
        if ket_idx is None:
            latex_list.append(R" + \ldots ")
        else:
            term = latex_terms[idx]
            ket = R"\texttt{" + ket_name(ket_idx) + "}"
            latex_list.append(Rf"{term}|{ket}\rangle &")

    return "".join(latex_list)


def _coeffs_to_latex_terms(coeffs: NDArray[np.complex128], decimals: int = 10) -> List[str]:
    """Convert a list of coefficients to latex formatted terms.

    The first non-zero term is treated differently. For this term a leading + is suppressed.

    Args:
        coeffs: List of coefficients to format
        decimals: Number of decimal places to round to (default: 10).
    Returns:
        List of formatted terms
    """
    first_term = True
    terms = []
    for coeff in coeffs:
        term = _coeff_to_latex_ket(coeff, first_term, decimals)
        if term is not None:
            first_term = False
        terms.append(term)
    return terms


def _coeff_to_latex_ket(raw_value: complex, first_coeff: bool, decimals: int = 10) -> str | None:
    """Convert a complex coefficient to latex code suitable for a ket expression.

    Args:
        raw_value: Value to convert
        first_coeff: If True then generate latex code for the first term in an expression
        decimals: Number of decimal places to round to (default: 10).
    Returns:
        String with latex code or None if no term is required
    """
    # Round to the specified number of decimals
    raw_value = np.around(raw_value, decimals=decimals)

    # If the value is zero then return None
    if np.abs(raw_value) == 0:
        return None

    # If the value has both real and imaginary parts, and the real part is negative, then extract the minus sign.
    # for example: +(-0.5+0.5j) -> -(0.5+0.5j)
    real_value = raw_value.real
    imag_value = raw_value.imag
    two_term_sign = "+"
    if np.sign(real_value) == -1 and imag_value != 0:
        two_term_sign = "-"
        raw_value = -raw_value

    # Convert to a sympy expression, then to latex
    value = sympy.nsimplify(raw_value, constants=(sympy.pi,), rational=False)
    latex_element = sympy.latex(value, full_prec=False)

    # Check if the value has more than one term
    two_term = real_value != 0 and imag_value != 0
    if isinstance(value, sympy.core.Add):
        # can happen for expressions like 1 + sqrt(2)
        two_term = True

    # If the value is 1 or -1 then suppress the coefficient
    if latex_element == "1":
        # If this is the first coefficient then suppress the leading + sign
        if first_coeff:
            return ""
        return "+"
    if latex_element == "-1":
        return "-"

    # If the value has more than one term, wrap it in parentheses
    if two_term:
        # If this is the first coefficient then suppress the leading + sign
        if first_coeff and two_term_sign == "+":
            return f"({latex_element})"
        return f"{two_term_sign}({latex_element})"

    # If this is not the first coefficient and the value is positive then add a leading + sign
    if not first_coeff and latex_element[0] != "-":
        return f"+{latex_element}"

    # Other normal case
    return latex_element


def _num_to_latex_ket(raw_value: complex, decimals: int = 15) -> str:
    """Convert a complex number to latex element.

    Args:
        raw_value: Value to convert
        decimals: Number of decimal places to round to (default: 15).

    Returns:
        str: latex element representing the value
    """
    raw_value = np.around(raw_value, decimals=decimals)
    value = sympy.nsimplify(raw_value, constants=(sympy.pi,), rational=False)
    return sympy.latex(value, full_prec=False)


def _latex_line_break(latex_code: str, output_length: int = 2) -> str:
    """Split latex string into several lines, so that each line has 2^output_length terms.

    Args:
        latex_code (str): latex source code
        output_length (int, optional): 2^output_length = number of terms in each line. Defaults to 2(= 4 terms/line).

    Returns:
        str: result latex code
    """
    latex_code = latex_code.replace("|", "&|")
    max_term = 2**output_length
    num_of_term = latex_code.count(R"\rangle")
    line_break = R"\\ &&"
    for i in range(max_term, num_of_term, max_term):
        new_line_index = find_nth_substring(latex_code, R"\rangle", i) + 7
        latex_code = latex_code[:new_line_index] + line_break + latex_code[new_line_index + 2 :]

    latex_code = latex_code.replace("&+", R"\;+&\;")
    latex_code = latex_code.replace("&-", R"\;-&\;")
    return latex_code


def _latex_code_to_img(latex_code: str) -> Image.Image:
    """Render latex code to an image,
    modified from qiskit.visualization.circuit.circuit_visualization._latex_circuit_drawer

    Args:
        latex_code: latex code to convert

    Returns:
        PIL.Image: an in-memory representation of the circuit diagram

    Raises:
        MissingOptionalLibraryError: if pillow, pdflatex, or poppler are not installed
        VisualizationError: if one of the conversion utilities failed for some internal or
            file-access reason.
    """
    header = R"""
\documentclass[border=2px]{standalone}

\usepackage{amsmath}
\usepackage{graphicx}

\begin{document}
"""
    footer = R"\end{document}"
    latex_source = header + latex_code + footer

    tmpfilename = "latex_source"
    with tempfile.TemporaryDirectory() as tmpdirname:
        tmppath = os.path.join(tmpdirname, tmpfilename + ".tex")

        with open(tmppath, "w") as tmpfile:
            tmpfile.write(latex_source)

        try:
            subprocess.run(
                [
                    "pdflatex",
                    "-halt-on-error",
                    f"-output-directory={tmpdirname}",
                    f"{tmpfilename + '.tex'}",
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.DEVNULL,
                check=True,
            )
        except OSError as exc:
            # OSError should generally not occur, because it's usually only triggered if `pdflatex`
            # doesn't exist as a command, but we've already checked that.
            raise VisualizationError("`pdflatex` command could not be run.") from exc
        except subprocess.CalledProcessError as exc:
            with open("latex_error.log", "wb") as error_file:
                error_file.write(exc.stdout)
            raise VisualizationError("`pdflatex` call did not succeed: see `latex_error.log`.") from exc
        base = os.path.join(tmpdirname, tmpfilename)
        try:
            subprocess.run(
                ["pdftocairo", "-singlefile", "-png", "-q", base + ".pdf", base],
                check=True,
            )
        except (OSError, subprocess.CalledProcessError) as exc:
            message = "`pdftocairo` failed to produce an image."
            raise VisualizationError(message) from exc
        image = Image.open(base + ".png")
        image = trim_image(image)
        return image



"""Module to store, manipulate and visualize quantum state vector."""

from __future__ import annotations

import copy
import itertools
import re
from typing import Iterable, List, Tuple

import numpy as np
from IPython.display import Latex
from numpy.typing import NDArray
from qiskit import QuantumCircuit
from qiskit.circuit import Instruction
from qiskit.exceptions import QiskitError
from qiskit.quantum_info import Statevector
from scipy import stats

from qiskit_extension import latex_drawer
from qiskit_extension.ket import Ket


class StateVector2(Statevector):
    """An extended class of Statevector from Qiskit:
    https://qiskit.org/documentation/stubs/qiskit.quantum_info.Statevector.html

    Args:
        data (np.ndarray | list | Statevector | QuantumCircuit | Instruction):
            Data from which the statevector can be constructed. This can be either a complex
            vector, another statevector, a ``QuantumCircuit`` or ``Instruction``.
            If the data is a circuit or instruction, the statevector is constructed by assuming that
            all qubits are initialized to the zero state.
        dims (int | tuple | list, optional): The subsystem dimension of the state.
    """

    def __init__(self, data, dims: int | Iterable[int] | None = None) -> None:
        if not isinstance(data, np.ndarray | list | Statevector | QuantumCircuit | Instruction):
            raise QiskitError("Input only accepts np.ndarray | list | Statevector | QuantumCircuit | Instruction")

        # REVERSE the order of qubits to fit textbook notation
        if isinstance(data, QuantumCircuit):
            data = data.reverse_bits()
        elif isinstance(data, Instruction):
            data = Statevector(data).reverse_qargs()

        self._data: NDArray[np.complex128]
        super().__init__(data, dims)
        self._num_of_qubit = int(np.log2(len(self._data)))
        self._basis = ["z"] * self._num_of_qubit

    @classmethod
    def __init__with_basis(
        cls, data, dims: int | Iterable[int] | None = None, basis: List[str] | str = []
    ) -> StateVector2:
        """Initialize statevector with basis."""
        state_vector = cls(data, dims)
        state_vector.basis = list(basis) + ["z"] * (state_vector._num_of_qubit - len(basis))
        return state_vector

    def __repr__(self) -> str:
        """Return the official string representation of a Statevector."""
        vector_str = np.array2string(self.to_matrix(), separator=", ")
        return f"Statevector:\n{vector_str}"

    @property
    def data(self) -> NDArray[np.complex128]:
        return self._data.copy()

    @property
    def basis(self) -> List[str]:
        return self._basis.copy()

    @data.setter
    def data(self, data: NDArray[np.complex128]):
        self._data = data.copy()

    @basis.setter
    def basis(self, basis: List[str]):
        self._basis = basis.copy()

    def copy(self) -> StateVector2:
        """Return a copy of the current statevector."""
        return copy.deepcopy(self)

    def to_matrix(self) -> NDArray[np.complex128]:
        """Return matrix form of statevector"""
        clone = self.copy()
        clone._basis_convert("z" * self._num_of_qubit)
        vector = clone._data
        matrix = vector[np.newaxis].T  # type: ignore
        return matrix

    def entropy(self, state: StateVector2 | Statevector | None = None) -> float:
        """Return entropy of input statevector."""
        if state is None:
            state = self
        return stats.entropy(state.probabilities(), base=2)  # type: ignore

    def evolve(self, circ: QuantumCircuit | NDArray, qargs: List[int] | None = None) -> StateVector2:
        """Evolve self(statevector) by a quantum circuit or matrix.
        return the evolved state as a new statevector. (self's state will not be changed)

        Args:
            circ (QuantumCircuit): The circ to evolve by.
            qargs (List[int], optional): A list of subsystem positions to apply the operator on.

        Returns:
            StateVector2: The new evolved statevector.
        """
        clone = self.copy()
        original_basis = clone.basis
        clone._basis_convert(["z"] * self._num_of_qubit)
        new_state = clone._evolve(circ, qargs)
        new_state._basis_convert(original_basis)
        return new_state

    def _evolve(self, circ: QuantumCircuit | NDArray, qargs: List[int] | None = None) -> StateVector2:
        """Evolve self(z-basis statevector) by a quantum circuit or matrix.
        return the evolved state as a new statevector. (self's state will not be changed)

        Args:
            circ (QuantumCircuit): The object to evolve by.
            qargs (List[int], optional): A list of subsystem positions to apply the operator on.

        Returns:
            StateVector2: The new evolved statevector.
        """
        if not isinstance(circ, QuantumCircuit | np.ndarray):
            raise QiskitError("Input is not a QuantumCircuit.")
        if isinstance(circ, QuantumCircuit):
            circ = circ.reverse_bits()  # REVERSE the order of qubits to fit textbook notation
        evolve_data = super().evolve(circ, qargs).data
        return StateVector2.__init__with_basis(data=evolve_data, basis=self.basis)

    def measure(
        self, measure: List[int] | int | str, basis: List[str] | str | None = None, shot=100
    ) -> List[StateVector2]:
        """measure statevector

        Args:
            measure (List[int] | int | str): which qubits to measure
            basis (List[str] | str | None, optional): measure in what basis. Defaults to current stored basis.
            shot (int, optional): number of shots. Defaults to 100.

        Returns:
            List[StateVector2]: list of statevector after measurement
        """
        states = self._measure(measure, basis, shot)[1]
        for state in states:
            if state is not None:
                state._basis_convert(self.basis)
        return states

    def _measure(
        self, measure: List[int] | int | str, basis: List[str] | str | None = None, shot=100
    ) -> Tuple[List[StateVector2], List[StateVector2]]:
        """measure statevector

        Args:
            measure (List[int] | int | str): which qubits to measure
            basis (List[str] | str, optional): measure in what basis. Defaults to current stored basis.
            shot (int, optional): number of shots. Defaults to 100.

        Returns:
            Tuple[List[StateVector2], List[StateVector2]]: list of measured statevector,
                list of remained statevector after measurement
        """
        clone = self.copy()

        if isinstance(measure, str):
            measure = [int(char) for char in measure]
        elif isinstance(measure, int):
            measure = [measure]
        if basis is not None:
            convert_basis = clone.basis
            for i in range(len(basis)):
                convert_basis[measure[i]] = basis[i]
            clone._basis_convert(convert_basis)

        measure_state_list = [None] * 2 ** len(measure)  # crate empty list for saving measure state
        remain_state_list = [None] * 2 ** len(measure)  # crate empty list for saving remain state
        for _ in range(shot):
            measure_ket: str
            remain_state: Statevector
            # REVERSE measure bit number to fit qiskit statevector measure notation
            reverse_measure = [clone._num_of_qubit - 1 - i for i in measure]
            measure_ket, remain_state = Statevector(clone.data).measure(qargs=reverse_measure)  # type: ignore
            measure_ket = measure_ket[::-1]  # REVERSE the order of qubits to fit textbook notation
            if measure_state_list[int(measure_ket, 2)] is None:
                basis = clone.basis
                measure_basis = ""
                for i in measure:
                    measure_basis += basis[i]

                basis_convert_measure_ket = ""
                for b, k in zip(measure_basis, measure_ket):
                    match b:
                        case "z":
                            basis_convert_measure_ket += Ket.z1 if int(k) else Ket.z0
                        case "x":
                            basis_convert_measure_ket += Ket.x1 if int(k) else Ket.x0
                        case "y":
                            basis_convert_measure_ket += Ket.y1 if int(k) else Ket.y0

                measure_state_list[int(measure_ket, 2)] = clone._from_label(basis_convert_measure_ket, to_z_basis=False)
                remain_state_list[int(measure_ket, 2)] = StateVector2.__init__with_basis(
                    data=remain_state.data, basis=clone.basis
                )
        return (measure_state_list, remain_state_list)

    def show_matrix(self):
        """show the matrix form of statevector in LaTeX"""
        return latex_drawer.matrix_to_latex(self.to_matrix())

    def show_state(
        self,
        basis: List[str] | str | None = None,
        hide: List[int] | str = [],
        output_length: int = 2,
        output: str = "latex",
    ):
        """visualize statevector

        Args:
            basis (List[str] | str, optional): convert statevector's basis to input basis.
                Defaults to skip basis convert.
            hide (List[int] | str, optional): hide qubits. Default to show all qubits.
            output_length (int, optional): 2^output_length = number of terms in each line. Defaults to 2(= 4 terms/line)
            output (str, optional): visualization method. Defaults to "latex".

        Returns:
            matplotlib.Figure or str or TextMatrix or IPython.display.Latex: visualization output
        """
        clone = self.copy()
        if basis is not None:
            clone._basis_convert(basis)

        match output:
            case "text":
                return print(clone)
            case "latex":
                return latex_drawer.state_to_latex(clone, hide, output_length)
            case "latex_source":
                return latex_drawer.state_to_latex(clone, hide, output_length, source=True)
            case _:
                return clone.draw(output=output)

    def show_measure(
        self,
        measure: List[int] | int | str,
        basis: List[str] | str | None = None,
        hide: List[int] | str = [],
        output_length: int = 2,
        remaining_basis: List[str] | str | None = None,
    ) -> Latex:
        """visualize measure result

        Args:
            measure (List[int] | int | str): qubits to measure
            basis (List[int] | str, optional): measure in what basis. Defaults to z basis.
            hide (List[int] | str, optional): hide qubits. Default to show all qubits.
            output_length (int, optional): 2^output_length = number of terms in each line. Defaults to 2(= 4 terms/line)
            remaining_basis (List[str] | str, optional): change statevector's basis to input basis before measure.

        Returns:
            Latex: visualization output
        """
        if isinstance(hide, str):
            hide = [int(char) for char in hide]
        elif isinstance(hide, int):
            hide = [hide]

        if isinstance(measure, str):
            measure = [int(char) for char in measure]
        elif isinstance(measure, int):
            measure = [measure]

        clone = self.copy()
        if remaining_basis is not None:
            remaining_basis = list(remaining_basis) + ["-"] * (self._num_of_qubit - len(remaining_basis))
            if basis is None:
                basis = ["z"] * len(measure)
            for i, b in zip(measure, basis):
                remaining_basis[i] = b
            clone._basis_convert(remaining_basis)
        result = clone._measure(measure=measure, basis=basis, shot=4 * 2 ** len(measure))
        return latex_drawer.measure_result_to_latex(result, measure + hide, output_length)  # type: ignore

    @classmethod
    def from_int(cls, i: int, dims: int | Iterable[int]) -> StateVector2:
        """Return a computational basis statevector.

        Args:
            i (int): the basis state element.
            dims (int | Iterable[int]): The subsystem dimensions of the statevector

        Returns:
            StateVector2: The statevector object.
        """
        state = super().from_int(i, dims)
        return cls(state.data, dims)

    @classmethod
    def from_instruction(cls, instruction):
        """Return the output statevector of an instruction.

        The statevector is initialized in the state :math:`|{0,\\ldots,0}\\rangle` of the
        same number of qubits as the input instruction or circuit, evolved
        by the input instruction, and the output statevector returned.

        Args:
            instruction (qiskit.circuit.Instruction or QuantumCircuit): instruction or circuit

        Returns:
            Statevector: The final statevector.

        Raises:
            QiskitError: if the instruction contains invalid instructions for
                         the statevector simulation.
        """
        if not isinstance(instruction, QuantumCircuit | Instruction):
            raise QiskitError("Input is not a valid instruction or circuit.")
        return cls(instruction)

    @classmethod
    def from_label(cls, *args: str | Tuple[complex, str]) -> StateVector2:
        """Create a state vector from input coefficient and label string.

        Example:
            from_label("0", "1") = (|0> + |1>)/√2,
            from_label("00", "01", "10", "11") = (|00> + |01> + |10> + |11>)/2 = |++>,
            from_label( (2**0.5,"0") , "+" , (-1,"-") ) = ( √2|0> + |+> - |-> )/2 = |+>

        Args:
            args (str | Tuple[complex, str]): The input label string or tuple of coefficient and label string.

        Returns:
            StateVector2: The statevector object.

        Raises:
            QiskitError: if labels contain invalid characters or labels have different number of qubits.
        """
        coeffs: List[complex] = []
        labels: List[str] = []
        for i, arg in enumerate(args):
            if isinstance(arg, tuple):
                coeffs.append(arg[0])
                labels.append(arg[1])
            else:
                coeffs.append(1.0)
                labels.append(arg)
            if not Ket.check_valid(labels[i]):
                raise QiskitError("Invalid label string.")
            if len(labels[0]) != len(labels[i]):
                raise QiskitError("Each label's number of qubits must be the same.")

        state: StateVector2 = cls._from_label(labels[0]) * coeffs[0]
        for coeff, label in zip(coeffs[1:], labels[1:]):
            state += cls._from_label(label) * coeff

        if type(state) is not StateVector2:
            raise QiskitError("Unexpected error.")
        state /= super(type(state), state).trace() ** 0.5
        return state

    @classmethod
    def _from_label(cls, label: str, to_z_basis: bool = True) -> StateVector2:
        if to_z_basis:
            label = label.replace(Ket.z0, "0")
            label = label.replace(Ket.x0, "+")
            label = label.replace(Ket.y0, "r")
            label = label.replace(Ket.z1, "1")
            label = label.replace(Ket.x1, "-")
            label = label.replace(Ket.y1, "l")
            state = super().from_label(label)
            return cls(state.data)

        basis = ""
        for ket in label:
            match ket:
                case Ket.z0 | Ket.z1:
                    basis += "z"
                case Ket.x0 | Ket.x1:
                    basis += "x"
                case Ket.y0 | Ket.y1:
                    basis += "y"

        label = label.replace(Ket.z0, "0")
        label = label.replace(Ket.x0, "0")
        label = label.replace(Ket.y0, "0")
        label = label.replace(Ket.z1, "1")
        label = label.replace(Ket.x1, "1")
        label = label.replace(Ket.y1, "1")
        state = super().from_label(label)
        return cls.__init__with_basis(state.data, basis=basis)

    def expand(self, other):
        """Return the tensor product state other ⊗ self.

        Args:
            other (Statevector2): a quantum state object.

        Returns:
            Statevector: the tensor product state other ⊗ self.

        Raises:
            QiskitError: if other is not a quantum state.
        """
        clone = self.copy()
        clone._basis_convert("z" * self._num_of_qubit)
        other_clone = other.copy()
        other_clone._basis_convert("z" * other._num_of_qubit)
        state_expand = super(type(other_clone), other_clone).expand(clone)
        return StateVector2(state_expand.data)

    def _basis_convert(
        self,
        basis: List[str] | str = [],
        algorithm: str = "global",
    ) -> None:
        """Convert basis of statevector

        Args:
            basis (List[str] | str, optional):new basis. Defaults to auto choose basis with minimum entropy.
            algorithm (str, optional): if don't specify which basis to convert, (basis = "-")
                convert basis to basis with "local" or "global" minimum entropy. Defaults to "global".
        """
        # check if input is valid
        target_basis = list(basis) + ["-"] * (self._num_of_qubit - len(basis))
        del basis
        if re.match(R"^[\-xyz]+$", "".join(target_basis)) is None:
            raise QiskitError("Invalid basis.")

        # convert basis using QuantumCircuit
        auto_basis_index = []
        circ_convert = QuantumCircuit(self._num_of_qubit)
        for i in range(self._num_of_qubit):
            if target_basis[i] == self._basis[i]:
                continue
            if target_basis[i] != "-":
                self._xyz_convert_circ(self.basis[i], target_basis[i], circ_convert, i)
                self._basis[i] = target_basis[i]
            else:
                auto_basis_index.append(i)
        self._data = self._evolve(circ_convert)._data
        if not auto_basis_index:
            return

        # if user don't specify which basis to convert, convert basis to basis with minimum entropy
        match algorithm:
            case "global":
                if len(auto_basis_index) > 8:
                    print(
                        "Warning: global minimum entropy basis convert with more then 8 qubits might take a long time."
                    )
                new_basis = self._global_min_entropy_basis(auto_basis_index)
            case "local":
                new_basis = self._local_min_entropy_basis(auto_basis_index)
            case _:
                raise QiskitError("Invalid min_entropy_basis_find_method.")
        self._basis_convert(new_basis)

    def _global_min_entropy_basis(self, auto_basis_index: List[int]) -> List[str]:
        """find basis with global minimum entropy

        Args:
            auto_basis_index (List[int]): index of auto-choose-basis

        Returns:
            List[str]: basis with global minimum entropy
        """
        num_of_auto_basis = len(auto_basis_index)
        min_entropy = float("inf")
        min_basis = self.basis
        try_basis = self.basis
        for basis in itertools.product(["z", "x", "y"], repeat=num_of_auto_basis):  # type: ignore
            for i in range(num_of_auto_basis):
                try_basis[auto_basis_index[i]] = basis[i]
            try_state = self.copy()
            try_state._basis_convert(try_basis)
            if (entropy := try_state.entropy()) < min_entropy:
                min_entropy = entropy
                min_basis = try_basis.copy()
        return min_basis

    def _local_min_entropy_basis(self, auto_basis_index: List[int]) -> List[str]:
        """find basis with local minimum entropy

        Args:
            auto_basis_index (List[int]): index of auto-choose-basis

        Returns:
            List[str]: basis with local minimum entropy
        """
        # Step 1: Change all auto-choose-basis to y, e.g. [-, -, -, -] -> [z, z, z, z], calculate entropy
        # Step 2,3: Same as Step 1, but with x-basis and y-basis
        # Step 4: from Step 1 to 3, choose the basis with minimum entropy.
        clone_state = self.copy()

        min_entropy = float("inf")
        min_basis = clone_state.basis
        for basis in ["z", "x", "y"]:
            try_state = clone_state.copy()
            try_basis = try_state.basis
            for i in auto_basis_index:
                try_basis[i] = basis
            try_state._basis_convert(try_basis)
            if (entropy := try_state.entropy()) < min_entropy:
                min_entropy = entropy
                min_basis = try_basis
        clone_state._basis_convert(min_basis)

        # Step 1: Change the first auto-choose-basis to y, e.g. [-, -, -, -] -> [y, -, -, -], calculate entropy,
        # Step 2,3: Same as Step 1, but with x-basis and z-basis
        # Step 4: from Step 1 to 3, choose the basis with minimum entropy.
        # Step 5: Repeat Step 1 to 4 for the second auto-choose-basis, and so on. (greedy)
        # e.g. [-, -, -, -] -> [x, -, -, -] -> [x, z, -, -] -> [x, z, y, -] -> [x, z, y, z]
        min_entropy = float("inf")
        min_basis = clone_state.basis
        for i in auto_basis_index:
            try_basis = clone_state.basis
            for basis in ["y", "x", "z"]:
                try_basis[i] = basis
                try_state = clone_state.copy()
                try_state._basis_convert(try_basis)
                if (entropy_tmp := try_state.entropy()) < min_entropy:
                    min_entropy = entropy_tmp
                    min_basis[i] = basis
            clone_state._basis_convert(min_basis)
        return min_basis

    @staticmethod
    def _xyz_convert_circ(basis: str, new_basis: str, circ: QuantumCircuit, qubit: int) -> None:
        """Add the corresponding gate that converts different basis

        Args:
            basis (str): original basis
            new_basis (str): new basis
            circ (QuantumCircuit): quantum circuit to convert basis

        Returns:
            QuantumCircuit: quantum circuit to convert basis
        """
        if basis == new_basis:
            return
        basis += new_basis
        match basis:
            case "zx" | "xz":
                circ.h(qubit)
            case "zy":
                circ.sdg(qubit)
                circ.h(qubit)
            case "yz":
                circ.h(qubit)
                circ.s(qubit)
            case "xy":
                circ.h(qubit)
                circ.sdg(qubit)
                circ.h(qubit)
            case "yx":
                circ.h(qubit)
                circ.s(qubit)
                circ.h(qubit)
        return



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="QiskitExtension",
    version="0.1.3",
    author="HSIEH, LI-YU",
    author_email="cjh9027@smail.nchu.edu.tw",
    description="A Qiskit visualize extension",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    url="https://github.com/Slope86/QiskitExtension",
    packages=setuptools.find_packages(exclude=["experiments", "tests"]),
    package_data={"": ["*.ini"]},
    include_package_data=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=["qiskit[visualization] >= 0.22.3"],
    python_requires=">=3.10",
)



def find_nth_substring(string: str, sub_string, n) -> int:
    """Find the index of nth substring in a string.

    Example:
        >>> find_nth_substring('123abcabcabc', 'abc', 2)
        6

    Args:
        string (str): The string to search.
        sub_string (_type_): The substring to search for.
        n (_type_): The nth substring to find.

    Returns:
        int: The index of the nth substring.
    """
    sub_string_index = string.find(sub_string)
    while sub_string_index >= 0 and n > 1:
        sub_string_index = string.find(sub_string, sub_string_index + len(sub_string))
        n -= 1
    return sub_string_index


if __name__ == "__main__":
    print(find_nth_substring("123abcabcabc", "abc", 2))



"""Module to store and check the ket notation for the state vector"""

import configparser
import importlib.resources as pkg_resources
import re

from qiskit_extension import config


class _ConfigMeta(type):
    __z0: str = "0"
    __z1: str = "1"
    __x0: str = "+"
    __x1: str = "-"
    __y0: str = "i"
    __y1: str = "j"

    def __new__(cls, name, bases, namespace):
        config_parser = configparser.ConfigParser()
        with pkg_resources.path(config, "config.ini") as path:
            config_parser.read(path)
        for ket in config_parser["ket"]:
            if ket not in ["z0", "z1", "x0", "x1", "y0", "y1"]:
                print(f"Unknown basis [{ket}]. Basis should be z0, z1, x0, x1, y0, y1.")
                print("Please check the config.ini file, invalid user setting will be ignored and use default setting.")
                continue
            if len(config_parser["ket"][ket]) != 1:
                print(f"The ket notation should be a single character but [{config_parser['ket'][ket]}] is given.")
                print("Please check the config.ini file, invalid user setting will be ignored and use default setting.")
                continue
            namespace[f"__{ket}"] = config_parser["ket"][ket]

        return super().__new__(cls, name, bases, namespace)

    @property
    def z0(self) -> str:
        return self.__z0

    @property
    def z1(self) -> str:
        return self.__z1

    @property
    def x0(self) -> str:
        return self.__x0

    @property
    def x1(self) -> str:
        return self.__x1

    @property
    def y0(self) -> str:
        return self.__y0

    @property
    def y1(self) -> str:
        return self.__y1


class Ket(metaclass=_ConfigMeta):
    """Class to store and check the ket notation for the state vector"""

    @classmethod
    def check_valid(cls, label: str) -> bool:
        """
        Check if the input label is valid.
        """
        ket_regex = "^[" + cls.z0 + cls.z1 + cls.x0 + cls.x1 + cls.y0 + cls.y1 + "]+$"
        # escape the minus sign in the regex
        # e.g. if ket notation = |0>, |1>, |+>, |->, |i>, |j>, then ket_str = R"^[01+\-rj]+$"
        ket_regex = ket_regex.replace("-", R"\-")

        if re.match(ket_regex, label) is None:
            return False
        return True
